﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Drawing;

namespace FileTypeVbp
{
    public class VbpFile
    {
        public struct VbpFileHeader
        {
            public char[] Type;
            public uint Size;
            public uint Reserved;
            public uint OffBits;
        }

        public enum IH_Compression
        {
            BI_RGB = 0,
            BI_RLE8 = 1,
            BI_RLE4 = 2,
            BI_BITFIELDS = 3
        }

        public struct IH_ColorMask
        {
            public uint R;
            public uint G;
            public uint B;
            public uint Mask;
        }

        public struct VbpInformationHeader{
            public uint Size;
            public int Width;
            public int Height;
            public ushort Planes;
            public ushort BitCount;
            public IH_Compression Compression;
            public uint SizeImage;
            public int XPelsPerMeter;
            public int YPelsPerMeter;
            public uint ClrUsed;
            public uint ClrImportant;
            public IH_ColorMask ColorMask;
        }

        Dictionary<uint, uint> _colorTable8Bit = new Dictionary<uint, uint>();
        Dictionary<uint, uint> _colorTable5Bit = new Dictionary<uint, uint>();
               
        public VbpFile()
        {

            for (uint i = 0; i <= 255; i++)
            {
                _colorTable8Bit[i] = (uint)(i / 8.2);
                if (i > 0)
                {
                    if (_colorTable8Bit[i - 1] != _colorTable8Bit[i])
                    {
                        _colorTable5Bit[_colorTable8Bit[i]] = i;
                    }
                }
                else
                {
                    _colorTable5Bit[0] = 0;
                }
            }
        }

        public Bitmap ReadBitmap(Stream input)
        {
            Bitmap ret = new Bitmap(160,120);
            VbpFileHeader _fileHeader = ReadFileHeader(input);
            VbpInformationHeader _infoHeader = ReadInfoHeader(input);
            if (_infoHeader.Compression == IH_Compression.BI_BITFIELDS)
            {
                ret = ReadBitmapData(input, _fileHeader, _infoHeader);
            }
            
            return ret;
        }

        public void WriteBitmap(Bitmap input, Stream output)
        {
            WriteFileHeader(output, input);
            WriteInfoHeader(output, input);
            WriteBitmapData(output, input);
        }

        protected VbpFileHeader ReadFileHeader(Stream input)
        {
            input.Position = 0;
            byte[] imageData = new byte[14];
            input.Read(imageData, 0, 14);

            VbpFileHeader _fileHeader = new VbpFileHeader();

            _fileHeader.Type = new char[2];
            _fileHeader.Type[0] = (char)imageData[0];
            _fileHeader.Type[1] = (char)imageData[1];

            if ((_fileHeader.Type[0] != 'B') || (_fileHeader.Type[1] != 'M'))
            {
                throw new Exception("Wrong Filetype!");
            }
            _fileHeader.Size = BitConverter.ToUInt32(imageData, 2);
            _fileHeader.Reserved = BitConverter.ToUInt32(imageData, 6);
            _fileHeader.OffBits = BitConverter.ToUInt32(imageData, 10);

            return _fileHeader;
        }

        protected void WriteFileHeader(Stream output, Bitmap image)
        {
            WriteBytes(new char[] { 'B', 'M' }, 2, output);
            uint _size = (uint)(image.Width * image.Height * 2)+ 0x46;
            output.Write(BitConverter.GetBytes(_size), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x00000000), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x00000046), 0, 4);
        }


        protected VbpInformationHeader ReadInfoHeader(Stream input)
        {
            input.Position = 14;
            byte[] imageData = new byte[40];
            input.Read(imageData, 0, 40);

            VbpInformationHeader _infoHeader = new VbpInformationHeader();

            _infoHeader.Size = BitConverter.ToUInt32(imageData, 0);
            _infoHeader.Width = BitConverter.ToInt32(imageData, 4);
            _infoHeader.Height = BitConverter.ToInt32(imageData, 8);
            _infoHeader.Planes = BitConverter.ToUInt16(imageData, 12);
            _infoHeader.BitCount = BitConverter.ToUInt16(imageData, 14);
            uint _value = BitConverter.ToUInt32(imageData, 16);
            _infoHeader.Compression = (IH_Compression)_value;
            _infoHeader.SizeImage = BitConverter.ToUInt32(imageData, 20);
            _infoHeader.XPelsPerMeter = BitConverter.ToInt32(imageData, 24);
            _infoHeader.YPelsPerMeter = BitConverter.ToInt32(imageData, 28);
            _infoHeader.ClrUsed = BitConverter.ToUInt32(imageData, 32);
            _infoHeader.ClrImportant = BitConverter.ToUInt32(imageData, 36);

            if (_infoHeader.Compression == IH_Compression.BI_BITFIELDS)
            {
                input.Position = 54;
                imageData = new byte[16];
                input.Read(imageData, 0, 16);

                _infoHeader.ColorMask.R = BitConverter.ToUInt32(imageData, 0);
                _infoHeader.ColorMask.G = BitConverter.ToUInt32(imageData, 4);
                _infoHeader.ColorMask.B = BitConverter.ToUInt32(imageData, 8);
                _infoHeader.ColorMask.Mask = BitConverter.ToUInt32(imageData, 12);
            }
            else
            {
                throw new Exception("Wrong Filetype!");
            }

            return _infoHeader;
        }

        private void WriteInfoHeader(Stream output, Bitmap input)
        {
            output.Write(BitConverter.GetBytes((uint)0x00000038), 0, 4);
            output.Write(BitConverter.GetBytes(input.Width), 0, 4);
            output.Write(BitConverter.GetBytes(input.Height), 0, 4);
            output.Write(BitConverter.GetBytes((ushort)1), 0, 2);
            output.Write(BitConverter.GetBytes((ushort)16), 0, 2);
            output.Write(BitConverter.GetBytes((uint)IH_Compression.BI_BITFIELDS), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x00000062), 0, 4);
            output.Write(BitConverter.GetBytes((int) 0x00002e20), 0, 4);
            output.Write(BitConverter.GetBytes((int) 0x00002e20), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x00000000), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x00000000), 0, 4);

            output.Write(BitConverter.GetBytes((uint)0x00007c00), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x000003e0), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x0000001f), 0, 4);
            output.Write(BitConverter.GetBytes((uint)0x00008000), 0, 4);
        }

        protected Bitmap ReadBitmapData(Stream input, VbpFileHeader fileHeader, VbpInformationHeader infoHeader)
        {
            input.Position = fileHeader.OffBits;
            int _size = infoHeader.Width * infoHeader.Height * 2;
            byte[] imageData = new byte[_size];
            input.Read(imageData, 0, _size);

            Bitmap _bitmap = new Bitmap(infoHeader.Width, infoHeader.Height);

            int _shiftR = GetShift(infoHeader.ColorMask.R);
            int _shiftG = GetShift(infoHeader.ColorMask.G);
            int _shiftB = GetShift(infoHeader.ColorMask.B);

            int _count = 0;
            for (int _y = 0; _y < infoHeader.Height; _y++)
            {
                for (int _x = 0; _x < infoHeader.Width; _x++)
                {
                    ushort _pixel = BitConverter.ToUInt16(imageData, (_count++ *2));
                    ushort _r = (ushort)_colorTable5Bit[((_pixel & infoHeader.ColorMask.R) >> _shiftR)];
                    ushort _g = (ushort)_colorTable5Bit[((_pixel & infoHeader.ColorMask.G) >> _shiftG)];
                    ushort _b = (ushort)_colorTable5Bit[((_pixel & infoHeader.ColorMask.B) >> _shiftB)];
                    Color _color = Color.FromArgb(_r, _g, _b);
                    _bitmap.SetPixel(_x, _y, _color);
                    
                }
            }

            return _bitmap;
        }
        protected void WriteBitmapData(Stream output, Bitmap input)
        {
            int _shiftR = 10;
            int _shiftG = 5;
            int _shiftB = 0;

            for (int _y = 0; _y < input.Height; _y++)
            {
                for (int _x = 0; _x < input.Width; _x++)
                {
                    Color _color = input.GetPixel(_x, _y);
                    ushort _b = (ushort)((_colorTable8Bit[_color.B]) << _shiftB);
                    ushort _g = (ushort)((_colorTable8Bit[_color.G]) << _shiftG);
                    ushort _r = (ushort)((_colorTable8Bit[_color.R]) << _shiftR);
                    ushort _rgb = (ushort)(0x8000 | _r | _g | _b);
                    output.Write(BitConverter.GetBytes(_rgb), 0, 2);
                }
            }
        }

        protected int GetShift(uint mask)
        {
            int ret = 0;
            uint _r = mask;

            if ((_r & 0x1) == 0)
            {
                do
                {
                    ret++;
                    _r = _r >> 1;
                } while ((_r & 0x1) == 0);
            }

            return ret;
        }

        void WriteBytes(char[] values, int count, Stream output)
        {
            for (int i = 0; i < count; i++)
            {
                output.WriteByte((byte)values[i]);
            }
        }
    }
}
